##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2014 Torsten Duwe <duwe@suse.de>
## Copyright (C) 2014 Sebastien Bourdelin <sebastien.bourdelin@savoirfairelinux.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from collections import deque

class SamplerateError(Exception):
    pass
    
def bit2str(t):
    index = 0
    retval = ""
    for x in t:
        if(x == 1):
            retval = "1" + retval
        else:
            retval = "0" + retval
        index = index + 1
    return retval

def bit2val(t):
    index = 0
    retval = 0
    for x in t:
        if(x == 1):
            retval = retval + pow(2, index)
        index = index + 1
    return retval

class Decoder(srd.Decoder):
    api_version = 3
    id = 'induction'
    name = 'Induction'
    longname = 'Decoder for Induction Field'
    desc = 'Trying to decode Induction field.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = ['induction']
    channels = (
        {'id': 'data', 'name': 'Data', 'desc': 'Data line'},
    )
    annotations = (
        ('data', 'Data'),
        ('bit', 'Bit'),
    )
    annotation_rows = (
        ('data', 'Data', (0,)),
        ('bit', 'Bit', (1,)),
    )
    options = (
        { 'id': 'start_high_time', 'desc': 'Start High Time [ms]', 'default': 25 },
        { 'id': 'start_high_time_tol', 'desc': 'Start High Time Tol. +- [ms]', 'default': 10 },
        { 'id': 'bit_time_high', 'desc': 'Bit time high [us]', 'default': 5000 },
        { 'id': 'bit_time_high_tol', 'desc': 'Bit time high Time Tol. [us]', 'default': 1500 },
        { 'id': 'bit_time_low', 'desc': 'Bit time low [us]', 'default': 1000 },
        { 'id': 'bit_time_low_tol', 'desc': 'Bit time low Time Tol. [us]', 'default': 500 },
        { 'id': 'glitch_max', 'desc': 'Glitch Time Max. to filter out [ns]', 'default': 1000 },
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.samplerate = None
        self.last_samplenum = None
        self.chunks = 0
        self.level_changed = False
        self.last_t = None
        self.state = "WAITFORSTART1"
        self.bitcnt = 0
        self.bytestart = 0
        self.byte = ()

    def metadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def decode(self):
        if not self.samplerate:
            raise SamplerateError('Cannot decode without samplerate.')
        while True:
            pin = self.wait({0: 'e'})

            if not self.last_samplenum:
                self.last_samplenum = self.samplenum
                continue
                
            samples = self.samplenum - self.last_samplenum
            t = samples / self.samplerate
            
            if(t<self.options['glitch_max']*1e-9):
                continue
            
            if(pin[0] == 0):
            
                if(self.state == "WAITFORSTART1"):
                    startmin = (self.options['start_high_time'] - self.options['start_high_time_tol'])*1e-3
                    startmax = (self.options['start_high_time'] + self.options['start_high_time_tol'])*1e-3
                    
                    self.bytestart = 0
                    self.bitcnt = 0
                    self.byte = ()
                    
                    if(t > startmin and t < startmax):
                        self.state = "DECODEBITS"
                        self.bitcnt = 0
                        self.bytestart = self.samplenum
                elif(self.state == "DECODEBITS"):
                    
                    one_min = (self.options['bit_time_high'] - self.options['bit_time_high_tol'])*1e-6
                    one_max = (self.options['bit_time_high'] + self.options['bit_time_high_tol'])*1e-6
                    zero_min = (self.options['bit_time_low'] - self.options['bit_time_low_tol'])*1e-6
                    zero_max = (self.options['bit_time_low'] + self.options['bit_time_low_tol'])*1e-6
                    
                    bitval = "??"
                    
                    if(t > one_min and t < one_max):
                        self.byte = (1,) + self.byte
                        bitval = "1"

                    if(t > zero_min and t < zero_max):
                        if(bitval != "??"):
                            self.put(self.bytestart, self.samplenum, self.out_ann,
                                 [0, ["Bit ERROR: " + str(self.bitcnt)]])
                        else:
                            self.byte = (0,) + self.byte
                            bitval = "0"
                                 
                    if(bitval != "ERR" and bitval != "??"):
                        self.bitcnt = self.bitcnt + 1
                        if(self.bitcnt >= 33):                          
                            self.put(self.bytestart, self.samplenum, self.out_ann,
                                 [0, [hex(bit2val(self.byte))]])
                            self.put(self.bytestart, self.samplenum, self.out_ann,
                                 [1, [bit2str(self.byte)]])
                            
                            self.state = "WAITFORSTART1"
                    else:
                        self.state = "WAITFORSTART1"
                        self.put(self.bytestart+6, self.samplenum, self.out_ann,
                                 [0, ["Bit Unkown: " + str(self.bitcnt) + " " + str(t/1e-9) + "ns"]])
                        self.put(self.bytestart+6, self.samplenum, self.out_ann,
                                 [1, ["Bit Unkown: " + str(self.bitcnt) + " " + str(t/1e-9) + "ns"]])
                             
            self.last_t = t
            self.last_samplenum = self.samplenum
